from datetime import datetime
import logging
import copy

from ..exceptions import CTERAException
from ..convert import fromxmlstr, toxmlstr
from ..common import Device, delete_attrs
from ..lib import TempfileServices
from ..lib.storage import commonfs, synfs
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.edge')


class Config(BaseCommand):
    """ Edge Filer General Configuration APIs """

    def get_location(self):
        """
        Get the location of the Edge Filer

        :return str: The location of the Edge Filer
        """
        return self._edge.api.get('/config/device/location')

    def set_location(self, location):
        """
        Set the location of the Edge Filer

        :param str location: New location to set
        :return str: The new location
        """
        logger.info('Configuring device location. %s', {'location': location})
        return self._edge.api.put('/config/device/location', location)

    def get_hostname(self):
        """
        Get the hostname of the Edge Filer

        :return str: The hostname of the Edge Filer
        """
        return self._edge.api.get('/config/device/hostname')

    def set_hostname(self, hostname):
        """
        Set the hostname of the Edge Filer

        :param str hostname: New hostname to set
        :return str: The new hostname
        """
        logger.info('Configuring device hostname. %s', {'hostname': hostname})
        return self._edge.api.put('/config/device/hostname', hostname)

    def import_config(self, config, exclude=None):
        """
        Import the Edge Filer configuration

        :param str config: A string or a path to the Edge Filer configuration file
        :param list[str],optional delete_attrs:
         List of configuration properties to exclude from import
        """
        database = None
        if isinstance(config, Device):
            database = copy.deepcopy(config)
        elif isinstance(config, str):
            database = Config.load_config(config)

        if exclude:
            delete_attrs(database, exclude)

        path = commonfs.join(TempfileServices.mkdir(), f'{self._edge.session().address}.xml')
        synfs.overwrite(path, toxmlstr(database, True).encode('utf-8'))

        return self._import_configuration(path)

    def _import_configuration(self, path):
        commonfs.properties(path)
        logger.info('Importing Edge Filer configuration.')
        with open(path, 'rb') as fd:
            response = self._edge.api.form_data(
                '/config',
                dict(
                    name='import',
                    type='db',
                    config=fd
                )
            )
            logger.info('Imported Edge Filer configuration.')
        return response

    @staticmethod
    def load_config(config):
        """
        Load the Edge Filer configuration

        :param str config: A string or a path to the Edge Filer configuration file
        """
        data = None
        if commonfs.exists(config):
            logger.info('Reading the Edge Filer configuration from file. %s', {'path': config})
            with open(config, 'r', encoding='utf-8') as f:
                data = f.read()
        else:
            data = config

        database = fromxmlstr(data)
        if database:
            logger.info('Completed parsing the Edge Filer configuration. %s', {'firmware': database.firmware})
            return database
        message = "An error occurred attempting to parse the Edge Filer's configuration."
        logger.error(message)
        raise CTERAException(message)

    def export(self, destination=None):
        """
        Export the Edge Filer configuration

        :param str,optional destination:
         File destination, defaults to the default directory
        """
        default_filename = self._edge.host() + datetime.now().strftime('_%Y-%m-%dT%H_%M_%S') + '.xml'
        directory, filename = commonfs.generate_file_destination(destination, default_filename)
        logger.info('Exporting configuration. %s', {'host': self._edge.host()})
        handle = self._edge.api.handle('/export')
        filepath = synfs.write(directory, filename, handle)
        logger.info('Exported configuration. %s', {'filepath': filepath})
        return filepath

    def is_wizard_enabled(self):
        """
        Get the current configuration of the first time wizard

        :return bool: True if the first time wizard is enabled, else False
        """
        return self._edge.api.get('/config/gui/openFirstTimeWizard')

    def enable_wizard(self):
        """
        Enable the first time wizard
        """
        return self._set_wizard(True)

    def disable_wizard(self):
        """
        Disable the first time wizard
        """
        return self._set_wizard(False)

    def _set_wizard(self, state):
        logger.info('Disabling first time wizard')
        return self._edge.api.put('/config/gui/openFirstTimeWizard', state)



from abc import abstractmethod
import cterasdk.settings
from ...clients import clients
from ..services import Management
from ..endpoints import EndpointBuilder
from .. import authenticators
from ...lib.session.core import Session
from ...core import (
    activation, admins, antivirus, buckets, cli, cloudfs, connection, credentials,
    devices, directoryservice, domains, files, firmwares, groups, kms, licenses,
    login, logs, mail, messaging, plans, portals, reports, roles, servers, settings,
    setup, ssl, startup, storage_classes, syslog, tasks, templates, users,
)


class Clients:

    def __init__(self, core):
        self.ctera = core.default.clone(clients.Extended, EndpointBuilder.new(core.base, core.context))
        self.api = core.default.clone(clients.API, EndpointBuilder.new(core.base, core.context, '/api'))
        self.io = IO(core)


class IO:

    def __init__(self, core):
        self._folders = core.default.clone(clients.Folders, EndpointBuilder.new(core.base, core.context, '/folders/folders'))
        self._upload = core.default.clone(clients.Upload, EndpointBuilder.new(core.base, core.context, '/upload/folders'))
        self._webdav = core.default.clone(clients.WebDAV, EndpointBuilder.new(core.base, core.context, '/webdav'))

    @property
    def upload(self):
        return self._upload.upload

    @property
    def download(self):
        return self._webdav.download

    @property
    def download_zip(self):
        return self._folders.download_zip

    @property
    def builder(self):
        return self._webdav._builder  # pylint: disable=protected-access


class Portal(Management):  # pylint: disable=too-many-instance-attributes

    def __init__(self, host, port=None, https=True):
        super().__init__(host, port, https, None, cterasdk.settings.core.syn.settings)
        self._ctera_session = Session(self.host(), self.context)
        self._ctera_clients = Clients(self)
        self.activation = activation.Activation(self)
        self.admins = admins.Administrators(self)
        self.backups = files.Backups(self)
        self.cloudfs = cloudfs.CloudFS(self)
        self.credentials = credentials.Credentials(self)
        self.devices = devices.Devices(self)
        self.directoryservice = directoryservice.DirectoryService(self)
        self.domains = domains.Domains(self)
        self.files = files.CloudDrive(self)
        self.firmwares = firmwares.Firmwares(self)
        self.groups = groups.Groups(self)
        self.logs = logs.Logs(self)
        self.plans = plans.Plans(self)
        self.reports = reports.Reports(self)
        self.roles = roles.Roles(self)
        self.settings = settings.Settings(self)
        self.storage_classes = storage_classes.StorageClasses(self)
        self.tasks = tasks.Tasks(self)
        self.templates = templates.Templates(self)
        self.users = users.Users(self)

    @property
    def ctera(self):
        return self.clients.ctera

    @property
    def api(self):
        return self.clients.api

    @property
    def io(self):
        return self.clients.io

    @property
    @abstractmethod
    def context(self):
        return NotImplementedError("Subclass must implement the 'context' property")

    @property
    def _session_id_key(self):
        return 'JSESSIONID'

    def _authenticator(self, url):
        return authenticators.core(self.session(), url, self.context)

    @property
    def _login_object(self):
        return login.Login(self)

    def test(self):
        return connection.test(self)

    def public_info(self):
        return self.ctera.get('/public/publicInfo')

    @property
    def _omit_fields(self):
        return super()._omit_fields + ['activation', 'admins', 'cloudfs', 'credentials', 'devices', 'directoryservice', 'domains', 'files',
                                       'firmwares', 'groups', 'logs', 'plans', 'reports', 'roles', 'settings', 'tasks', 'templates',
                                       'users']


class GlobalAdmin(Portal):  # pylint: disable=too-many-instance-attributes

    def __init__(self, host, port=None, https=True):
        super().__init__(host, port, https)
        self.antivirus = antivirus.Antivirus(self)
        self.buckets = buckets.Buckets(self)
        self.cli = cli.CLI(self)
        self.kms = kms.KMS(self)
        self.licenses = licenses.Licenses(self)
        self.mail = mail.Mail(self)
        self.messaging = messaging.Messaging(self)
        self.portals = portals.Portals(self)
        self.servers = servers.Servers(self)
        self.setup = setup.Setup(self)
        self.ssl = ssl.SSL(self)
        self.startup = startup.Startup(self)
        self.syslog = syslog.Syslog(self)

    @property
    def context(self):
        return 'admin'

    def impersonate(self, username, tenant):
        """
        Impersonate a Portal user

        :param str username: Username
        :param str tenant: Tenant
        """
        ctera_ticket = self.users.generate_ticket(username, tenant)
        user = ServicesPortal(f'{tenant}.{self.settings.global_settings.dns_suffix}', self.port())
        user.sso(ctera_ticket)
        return user

    @property
    def _omit_fields(self):
        return super()._omit_fields + ['antivirus', 'buckets', 'cli', 'kms', 'licenses', 'mail', 'messaging', 'portals', 'servers',
                                       'setup', 'ssl', 'startup', 'syslog']


class ServicesPortal(Portal):

    @property
    def context(self):
        return 'ServicesPortal'

    def sso(self, ctera_ticket):
        """
        Login using a Portal ticket

        :param str ctera_ticket: SSO Ticket.
        """
        self._login_object.sso(ctera_ticket)
        self.session().start_session(self)
        self.api.web_session()



from .base import CTERAException


class SessionExpired(CTERAException):
    """Session expiration"""

    def __init__(self):
        super().__init__('Authentication error: Session expired. Please log in again.')


class NotLoggedIn(CTERAException):
    """No session"""

    def __init__(self):
        super().__init__('Authentication error: Not logged in.')


class ContextError(CTERAException):
    """API invocation context rrror"""

    def __init__(self, message):
        super().__init__(f'Context error: {message}.')



import asyncio
import logging
import functools


logger = logging.getLogger('cterasdk.common')


def execute_with_retries(retries=None, backoff=None, max_backoff=None):
    """
    A decorator that retries a function or coroutine upon exception, using exponential backoff.

    This decorator supports both synchronous and asynchronous functions.

    :param int retries: The maximum number of attempts before giving up.
    :param int backoff: The initial backoff delay in seconds.
    """
    def decorator(func):
        @functools.wraps(func)
        async def a_wrapper(*args, **kwargs):
            delay = backoff
            for try_num in range(retries):
                try:
                    logger.debug("Try %s out of %s of function: '%s'", try_num + 1, retries, func.__name__)
                    return await func(*args, **kwargs)
                except Exception as e:  # pylint: disable=broad-exception-caught
                    logger.debug("Try %s of %s of function '%s' failed. Backing off for %s second(s).", try_num + 1, retries,
                                 func.__name__, delay)
                    if try_num == retries - 1:
                        raise e
                    await asyncio.sleep(delay)
                    delay = min(max_backoff, delay * 2)
        return a_wrapper
    return decorator



import asyncio
from .errors import XMLHandler, JSONHandler
from .base import BaseClient, BaseResponse, run_threadsafe
from .common import Serializers, Deserializers
from . import async_requests, decorators
from ..common import Object


class AsyncClient(BaseClient):
    """Asynchronous Client"""

    @decorators.authenticated
    async def get(self, path, *, on_response=None, on_error=None, **kwargs):
        request = async_requests.GetRequest(self._builder(path), **kwargs)
        return await self.a_request(request, on_response=on_response, on_error=on_error)

    @decorators.authenticated
    async def put(self, path, data, *, data_serializer=None, on_response=None, on_error=None, **kwargs):
        request = async_requests.PutRequest(self._builder(path), data=data_serializer(data), **kwargs)
        return await self.a_request(request, on_response=on_response, on_error=on_error)

    @decorators.authenticated
    async def post(self, path, data, *, data_serializer=None, on_response=None, on_error=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=data_serializer(data), **kwargs)
        return await self.a_request(request, on_response=on_response, on_error=on_error)

    @decorators.authenticated
    async def form_data(self, path, data, *, on_response=None, on_error=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=Serializers.FormData(data), **kwargs)
        return await self.a_request(request, on_response=on_response, on_error=on_error)

    @decorators.authenticated
    async def delete(self, path, *, on_response=None, on_error=None, **kwargs):
        request = async_requests.DeleteRequest(self._builder(path), **kwargs)
        return await self.a_request(request, on_response=on_response, on_error=on_error)

    async def _request(self, request, *, on_response=None, on_error=None):
        on_response = on_response if on_response else AsyncResponse.new()
        response = await self._session.await_promise(self.join_headers(request), on_response=on_response)
        return response if response.ok else await on_error.a_accept(response)


class AsyncFolders(AsyncClient):

    async def download_zip(self, path, data, **kwargs):
        return await super().form_data(path, data, **kwargs)


class AsyncUpload(AsyncClient):

    async def upload(self, path, data, **kwargs):
        response = await super().form_data(path, data, **kwargs)
        return await response.xml()


class AsyncWebDAV(AsyncClient):
    """WebDAV"""

    async def download(self, path, **kwargs):
        return await super().get(path, on_error=XMLHandler(), **kwargs)

    @decorators.authenticated
    async def propfind(self, path, depth, **kwargs):
        request = async_requests.PropfindRequest(self._builder(path), headers={'depth': str(depth)}, **kwargs)
        response = await self.a_request(request, on_error=XMLHandler())
        return await response.dav()

    @decorators.authenticated
    async def mkcol(self, path, **kwargs):
        request = async_requests.MkcolRequest(self._builder(path), **kwargs)
        response = await self.a_request(request, on_error=XMLHandler())
        return await response.text()

    @decorators.authenticated
    async def copy(self, source, destination, *, overwrite=False, **kwargs):
        request = async_requests.CopyRequest(self._builder(source), headers=self._webdav_headers(destination, overwrite), **kwargs)
        response = await self.a_request(request, on_error=XMLHandler())
        return await response.xml()

    @decorators.authenticated
    async def move(self, source, destination, *, overwrite=False, **kwargs):
        request = async_requests.MoveRequest(self._builder(source), headers=self._webdav_headers(destination, overwrite), **kwargs)
        response = await self.a_request(request, on_error=XMLHandler())
        return await response.xml()

    async def delete(self, path, **kwargs):  # pylint: disable=arguments-differ
        response = await super().delete(path, on_error=XMLHandler(), **kwargs)
        return await response.text()

    def _webdav_headers(self, destination, overwrite):
        return {
            'Destination': self._builder(destination),
            'Overwrite': 'T' if overwrite is True else 'F'
        }


class AsyncJSON(AsyncClient):

    def __init__(self, builder=None, session=None, settings=None, authenticator=None):
        super().__init__(builder, session, settings, authenticator)
        self.headers.persist_headers({'Content-Type': 'application/json'})

    async def get(self, path, **kwargs):
        response = await super().get(path, on_error=JSONHandler(), **kwargs)
        return await response.json()

    async def put(self, path, data, **kwargs):
        response = await super().put(path, data, data_serializer=Serializers.JSON, on_error=JSONHandler(), **kwargs)
        return await response.json()

    async def post(self, path, data, **kwargs):
        response = await super().post(path, data, data_serializer=Serializers.JSON, on_error=JSONHandler(), **kwargs)
        return await response.json()

    async def delete(self, path, **kwargs):
        response = await super().delete(path, on_error=JSONHandler(), **kwargs)
        return await response.json()


class AsyncXML(AsyncClient):

    async def get(self, path, **kwargs):
        response = await super().get(path, on_error=XMLHandler(), **kwargs)
        return await response.xml()

    async def put(self, path, data, **kwargs):
        response = await super().put(path, data, data_serializer=Serializers.XML, on_error=XMLHandler(), **kwargs)
        return await response.xml()

    async def post(self, path, data, **kwargs):
        response = await super().post(path, data, data_serializer=Serializers.XML, on_error=XMLHandler(), **kwargs)
        return await response.xml()

    async def delete(self, path, **kwargs):
        response = await super().delete(path, on_error=XMLHandler(), **kwargs)
        return await response.xml()


class AsyncExtended(AsyncXML):
    """CTERA Schema"""

    async def get_multi(self, path, paths, **kwargs):
        return await self.database(path, 'get-multi', paths, **kwargs)

    async def execute(self, path, name, param=None, **kwargs):  # schema method
        return await self._execute(path, 'user-defined', name, param, **kwargs)

    async def database(self, path, name, param=None, **kwargs):  # schema method
        return await self._execute(path, 'db', name, param, **kwargs)

    async def _execute(self, path, _type, name, param, **kwargs):
        data = Object()
        data.type = _type
        data.name = name
        data.param = param
        return await super().post(path, data, **kwargs)


class AsyncAPI(AsyncExtended):
    """CTERA Management API"""

    async def web_session(self):
        response = await AsyncClient.get(self, '/currentSession')
        self.headers.persist_response_header(response, 'X-CTERA-TOKEN')
        return await response.xml()

    async def defaults(self, classname):
        return self.get(f'/defaults/{classname}')


class AsyncResponse(BaseResponse):
    """Asynchronous Response Object"""

    async def a_iter_content(self, chunk_size=None):
        async for chunk in self._response.content.iter_chunked(chunk_size if chunk_size else 5120):
            yield chunk

    async def text(self):
        return await self._response.text()

    async def json(self):
        return Deserializers.JSON(await self._response.read())

    async def xml(self):
        return Deserializers.XML(await self._response.read())

    async def dav(self):
        return Deserializers.DAV(await self._response.read())

    @async_requests.decorate_stream_error
    async def read(self, n=-1):
        return await self._response.content.read(n)

    @staticmethod
    def new():
        async def new_response(response):
            return AsyncResponse(response)
        return new_response


class Client(BaseClient):
    """Synchronous Client"""

    @decorators.authenticated
    def handle(self, path, *, on_response=None, on_error=None, **kwargs):
        request = async_requests.GetRequest(self._builder(path), **kwargs)
        return self.request(request, on_response=on_response, on_error=on_error)

    @decorators.authenticated
    def get(self, path, *, on_response=None, on_error=None, **kwargs):
        request = async_requests.GetRequest(self._builder(path), **kwargs)
        return self.request(request, on_response=on_response, on_error=on_error)

    @decorators.authenticated
    def put(self, path, data, *, data_serializer=None, on_response=None, on_error=None, **kwargs):
        request = async_requests.PutRequest(self._builder(path), data=data_serializer(data), **kwargs)
        return self.request(request, on_response=on_response, on_error=on_error)

    @decorators.authenticated
    def post(self, path, data, *, data_serializer=None, on_response=None, on_error=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=data_serializer(data), **kwargs)
        return self.request(request, on_response=on_response, on_error=on_error)

    @decorators.authenticated
    def form_data(self, path, data, *, on_response=None, on_error=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=Serializers.FormData(data), **kwargs)
        return self.request(request, on_response=on_response, on_error=on_error)

    @decorators.authenticated
    def multipart(self, path, form, *, on_response=None, on_error=None, **kwargs):
        request = async_requests.PostRequest(self._builder(path), data=form.data, **kwargs)
        return self.request(request, on_response=on_response, on_error=on_error)

    @decorators.authenticated
    def delete(self, path, *, on_response=None, on_error=None, **kwargs):
        request = async_requests.DeleteRequest(self._builder(path), **kwargs)
        return self.request(request, on_response=on_response, on_error=on_error)

    def _request(self, request, *, on_response=None, on_error=None):
        on_response = on_response if on_response else SyncResponse.new()
        response = execute(self._session.await_promise, self.join_headers(request), on_response=on_response)
        return response if response.ok else on_error.accept(response)

    def close(self):  # pylint: disable=invalid-overridden-method
        return execute(super().close)


class Folders(Client):

    def download_zip(self, path, data, **kwargs):
        return super().form_data(path, data, **kwargs)


class Upload(Client):

    def upload(self, path, data, **kwargs):
        response = super().form_data(path, data, **kwargs)
        return response.xml()


class WebDAV(Client):

    def download(self, path, **kwargs):
        return super().handle(path, on_error=XMLHandler(), **kwargs)

    @decorators.authenticated
    def propfind(self, path, depth, **kwargs):
        request = async_requests.PropfindRequest(self._builder(path), headers={'depth': str(depth)}, **kwargs)
        response = self.request(request, on_error=XMLHandler())
        return response.dav()

    @decorators.authenticated
    def mkcol(self, path, **kwargs):
        request = async_requests.MkcolRequest(self._builder(path), **kwargs)
        response = self.request(request, on_error=XMLHandler())
        return response.xml()

    @decorators.authenticated
    def copy(self, source, destination, *, overwrite=False, **kwargs):
        request = async_requests.CopyRequest(self._builder(source), headers=self._webdav_headers(destination, overwrite), **kwargs)
        response = self.request(request, on_error=XMLHandler())
        return response.xml()

    @decorators.authenticated
    def move(self, source, destination, *, overwrite=False, **kwargs):
        request = async_requests.MoveRequest(self._builder(source), headers=self._webdav_headers(destination, overwrite), **kwargs)
        response = self.request(request, on_error=XMLHandler())
        return response.xml()

    def delete(self, path, **kwargs):  # pylint: disable=arguments-differ
        response = super().delete(path, on_error=XMLHandler(), **kwargs)
        return response.text()

    def _webdav_headers(self, destination, overwrite):
        return {
            'Destination': self._builder(destination),
            'Overwrite': 'T' if overwrite is True else 'F'
        }


class XML(Client):
    """XML Serializer and Deserializer"""

    def __init__(self, builder=None, session=None, settings=None, authenticator=None):
        super().__init__(builder, session, settings, authenticator)
        self._type = {'Content-Type': 'text/plain'}

    def get(self, path, **kwargs):
        response = super().get(path, on_error=XMLHandler(), **kwargs)
        return response.xml()

    def put(self, path, data, **kwargs):
        response = super().put(path, data, data_serializer=Serializers.XML, headers=self._type, on_error=XMLHandler(), **kwargs)
        return response.xml()

    def post(self, path, data, **kwargs):
        response = super().post(path, data, data_serializer=Serializers.XML, headers=self._type, on_error=XMLHandler(), **kwargs)
        return response.xml()

    def form_data(self, path, data, **kwargs):
        response = super().form_data(path, data, on_error=XMLHandler(), **kwargs)
        return response.xml()

    def delete(self, path, **kwargs):
        response = super().delete(path, on_error=XMLHandler(), **kwargs)
        return response.xml()


class JSON(Client):
    """JSON Serializer and Deserializer"""

    def get(self, path, **kwargs):
        response = super().get(path, on_error=JSONHandler(), **kwargs)
        return response.json()

    def put(self, path, data, **kwargs):
        response = super().put(path, data, data_serializer=Serializers.JSON, on_error=JSONHandler(), **kwargs)
        return response.json()

    def post(self, path, data, **kwargs):
        response = super().post(path, data, data_serializer=Serializers.JSON, on_error=JSONHandler(), **kwargs)
        return response.json()

    def delete(self, path, **kwargs):
        response = super().delete(path, on_error=JSONHandler(), **kwargs)
        return response.json()


class Extended(XML):
    """CTERA Schema"""

    def get_multi(self, path, paths, **kwargs):
        return self.database(path, 'get-multi', paths, **kwargs)

    def show_multi(self, path, paths, **kwargs):
        print(Serializers.JSON(self.get_multi(path, paths, **kwargs), no_log=False))

    def execute(self, path, name, param=None, **kwargs):  # schema method
        return self._execute(path, 'user-defined', name, param, **kwargs)

    def database(self, path, name, param=None, **kwargs):  # schema method
        return self._execute(path, 'db', name, param, **kwargs)

    def add(self, path, param=None, **kwargs):
        return self.database(path, 'add', param, **kwargs)

    def _execute(self, path, _type, name, param, **kwargs):
        data = Object()
        data.type = _type
        data.name = name
        data.param = param
        return super().post(path, data, **kwargs)


class API(Extended):
    """CTERA Management API"""

    def web_session(self):
        response = Client.get(self, '/currentSession')
        self.headers.persist_response_header(response, 'X-CTERA-TOKEN')
        return response.xml()

    def defaults(self, classname):
        return self.get(f'/defaults/{classname}')


class Migrate(JSON):
    """CTERA Migrate Service"""

    def login(self):
        response = Client.get(self, '/auth/user', on_error=JSONHandler())
        self.headers.persist_response_header(response, 'x-mt-x')
        return response.json()


event_loop = asyncio.new_event_loop()


def execute(target, *args, **kwargs):
    loop = asyncio.get_event_loop()

    if not loop.is_running():
        return loop.run_until_complete(target(*args, **kwargs))

    return run_threadsafe(event_loop, target, *args, **kwargs)


class SyncResponse(AsyncResponse):
    """Synchronous Response Object"""

    def iter_content(self, chunk_size=None):
        while True:
            try:
                yield execute(super().a_iter_content(chunk_size).__anext__)
            except StopAsyncIteration:
                break

    def text(self):  # pylint: disable=invalid-overridden-method
        return execute(super().text)

    def json(self):  # pylint: disable=invalid-overridden-method
        return execute(super().json)

    def xml(self):  # pylint: disable=invalid-overridden-method
        return execute(super().xml)

    def dav(self):  # pylint: disable=invalid-overridden-method
        return execute(super().dav)

    @staticmethod
    def new():
        async def new_response(response):
            return SyncResponse(response)
        return new_response



import re
import time
import logging
import asyncio
from datetime import datetime
from abc import ABC, abstractmethod

from ..common import Object
from ..common.enum import TaskRunningStatus
from ..exceptions.common import TaskWaitTimeoutError, AwaitableTaskException
from ..exceptions.transport import HTTPError


logger = logging.getLogger('cterasdk.common')


class BaseTask(Object):
    """
    Base Task

    :ivar int id: Task ID
    :ivar str name: Task name
    :ivar datetime.datetime start_time: Start time
    :ivar int elapsed_time: Elapsed time
    :ivar cterasdk.common.enum.TaskRunningStatus status: Status
    :ivar int percentage: Percentage
    :ivar datetime.datetime end_time: End time
    :ivar cterasdk.common.object.Object result: Task result
    """
    def __init__(self, task):
        super().__init__()
        self.id = task.id
        self.name = task.name
        self.start_time = datetime.fromisoformat(task.startTime)
        self.elapsed_time = task.elapsedTime
        self.status = task.status
        self.percentage = task.percentage
        self.end_time = datetime.fromisoformat(task.endTime) if task.endTime is not None else None
        self.result = task.result


class EdgeTask(BaseTask):
    """
    Edge Task

    :ivar str description: Task description
    """
    def __init__(self, task):
        super().__init__(task)
        self.description = task.description


class PortalTask(BaseTask):
    """
    Portal Task

    :ivar str progress_str: Task progress description
    :ivar int tenant: Tenant UID
    """
    def __init__(self, task):
        super().__init__(task)
        self.progress_str = task.progstring
        self.tenant = task.portalUid


class FilesystemTask(PortalTask):
    """
    Filesystem Task

    :ivar int files_processed: Files processed
    :ivar int bytes_processed: Bytes processed
    :ivar int total_files: Total files
    :ivar int total_bytes: Total bytes
    :ivar str error_type: Task error
    :ivar str file_in_progress: File in progress
    :ivar int user_uid: User UID
    :ivar cterasdk.common.object.Object cursor: Cursor
    """
    def __init__(self, task):
        super().__init__(task)
        self.files_processed = task.filesProcessed
        self.bytes_processed = task.bytesProcessed
        self.total_files = task.totalFiles
        self.total_bytes = task.totalBytes
        self.error_type = task.errorType
        self.file_in_progress = task.fileInProgress
        self.user_uid = task.userUid
        self.cursor = task.cursor


class AwaitableTask(ABC):

    def __init__(self, ctera, ref):
        self._ctera = ctera
        self._ref = self._task_reference(ref)

    @property
    def ref(self):
        return self._ref

    def wait(self, timeout=None, poll_interval=None):
        """
        Wait until the given task is complete, or until the timeout expires.

        If a positive `timeout` (in seconds) is specified, this function will block
        only up to that duration. If `timeout` is None or non-positive, it will
        wait indefinitely until the task completes.

        :param float,optional timeout: Raise exception in the event of a timeout
        :param float,optional poll_interval: Poll interval, defaults to 1 second
        """
        try:
            return _synchronous_wait(self, timeout, poll_interval)
        except HTTPError as error:
            raise AwaitableTaskException(f"An error occurred while retrieving the status of Task '{self.ref}'.",
                                         self._ctera.tasks.awaitable_task(self.ref)) from error

    async def a_wait(self, timeout=None, poll_interval=None):
        """
        Wait until the given task is complete, or until the timeout expires.

        If a positive `timeout` (in seconds) is specified, this function will block
        only up to that duration. If `timeout` is None or non-positive, it will
        wait indefinitely until the task completes.

        :param float,optional timeout: Raise exception in the event of a timeout
        :param float,optional poll_interval: Poll interval, defaults to 1 second
        """
        try:
            return await _asynchronous_wait(self, timeout, poll_interval)
        except HTTPError as error:
            raise AwaitableTaskException(f"An error occurred while retrieving the status of Task '{self.ref}'.",
                                         self._ctera.tasks.awaitable_task(self.ref)) from error

    @abstractmethod
    def _task_reference(self, ref):
        raise NotImplementedError("Subclass must implement the '_task_reference' function.")

    @abstractmethod
    def _task_status_object(self, task):
        raise NotImplementedError("Subclass must implement the '_task_status_object' function.")

    @abstractmethod
    def status(self):
        raise NotImplementedError("Subclass must implement the 'status' function.")

    @abstractmethod
    async def a_status(self):
        raise NotImplementedError("Subclass must implement the 'a_status' function.")

    def __str__(self):
        return self._ref

    def __repr__(self):
        return str(self)


class AwaitableEdgeTask(AwaitableTask):
    """
    Awaitable Edge Filer Task Object
    """
    def _task_reference(self, ref):
        uid = None
        if isinstance(ref, int):
            uid = str(ref)
        elif isinstance(ref, str):
            match = re.search('[1-9][0-9]*', ref)
            if match is not None:
                start, end = match.span()
                uid = ref[start: end]
        if uid is not None:
            return '/proc/bgtasks/' + uid
        logger.error('Failed to parse task identifier from reference: %s', ref)
        raise ValueError(f'Failed to parse task identifier from reference: {ref}')

    def _task_status_object(self, task):
        return EdgeTask(task)

    def status(self):
        """
        Synchronous function to retrieve task status.
        """
        return self._task_status_object(self._ctera.api.get(self._ref))

    async def a_status(self):
        """
        Asynchronous function to retrieve task status.
        """
        return self._task_status_object(await self._ctera.api.get(self._ref))


class AwaitablePortalTask(AwaitableTask):
    """
    Awaitable Portal Task Object
    """
    def _task_reference(self, ref):
        match = re.search('servers/[^/]*/bgTasks/[1-9][0-9]*$', ref)
        if not match:
            logger.error('Failed to parse task identifier from reference: %s', ref)
            raise ValueError(f'Failed to parse task identifier from reference: {ref}')
        return match.group(0)

    def _task_status_object(self, task):
        if task._classname == 'FileManagerBgTask':  # pylint: disable=protected-access
            return FilesystemTask(task)
        return PortalTask(task)

    def status(self):
        """
        Synchronous function to retrieve task status.
        """
        if self._ctera.session().in_tenant_context():
            return self._task_status_object(self._ctera.api.execute('', 'getTaskStatus', self._ref))
        return self._task_status_object(self._ctera.api.get(f'{self._ref}'))

    async def a_status(self):
        """
        Asynchronous function to retrieve task status.
        """
        if self._ctera.session().in_tenant_context():
            return self._task_status_object(await self._ctera.v1.api.execute('', 'getTaskStatus', self._ref))
        return self._task_status_object(await self._ctera.v1.api.get(f'{self._ref}'))


def _before_wait(timeout, poll_interval):
    timeout_at = None

    if timeout is not None:
        if not isinstance(timeout, (int, float)):
            raise ValueError('Timeout must be a positive int or float.')
        timeout_at = time.time() + timeout

    if poll_interval is not None and isinstance(poll_interval, (int, float)):
        raise ValueError('Poll interval must be a positive int or float.')
    poll_interval = poll_interval if poll_interval is not None else 1

    return timeout_at, poll_interval


def _synchronous_wait(awaitable_task, timeout=None, poll_interval=None):
    timeout_at, poll_interval = _before_wait(timeout, poll_interval)
    while True:
        task = awaitable_task.status()
        if task.status in [
            TaskRunningStatus.Completed,
            TaskRunningStatus.Disabled,
            TaskRunningStatus.Stopped,
            TaskRunningStatus.Warnings,
            TaskRunningStatus.Failed,
        ]:
            return task
        if timeout_at is not None and time.time() > timeout_at:
            raise TaskWaitTimeoutError(timeout, task)
        time.sleep(poll_interval)


async def _asynchronous_wait(awaitable_task, timeout=None, poll_interval=None):
    timeout_at, poll_interval = _before_wait(timeout, poll_interval)
    while True:
        task = await awaitable_task.a_status()
        if task.status in [
            TaskRunningStatus.Completed,
            TaskRunningStatus.Disabled,
            TaskRunningStatus.Stopped,
            TaskRunningStatus.Warnings,
            TaskRunningStatus.Failed,
        ]:
            return task
        if timeout_at is not None and time.time() > timeout_at:
            raise TaskWaitTimeoutError(timeout, task)
        await asyncio.sleep(poll_interval)



import logging
from datetime import datetime
from contextlib import contextmanager
from pathlib import Path
from ..common import Object
from ..objects.uri import unquote
from . import common
from ..exceptions.transport import HTTPError
from ..exceptions.io import ResourceExistsError, RestrictedPathError, RemoteStorageError


logger = logging.getLogger('cterasdk.edge')


class EdgePath(common.BasePath):
    """Path for CTERA Edge Filer"""

    def __init__(self, scope, reference):
        """
        Initialize a CTERA Edge Filer Path.

        :param str scope: Scope.
        :param str reference: Reference.
        """
        if isinstance(reference, Object):
            super().__init__(scope, reference.path)
        elif isinstance(reference, str):
            super().__init__(scope, reference)
        elif reference is None:
            super().__init__(scope, '')
        else:
            message = 'Path validation failed: ensure the path exists and is correctly formatted.'
            logger.error(message)
            raise ValueError(message)

    @staticmethod
    def instance(scope, reference):
        return EdgePath(scope, reference)


def fetch_reference(href):
    namespace = '/localFiles'
    return unquote(href[href.index(namespace)+len(namespace) + 1:])


def format_listdir_response(parent, response):
    entries = []
    for e in response:
        path = fetch_reference(e.href)
        if path and parent != path:
            is_dir = e.getcontenttype == 'httpd/unix-directory'
            param = Object(
                path=path,
                name=Path(path).name,
                is_dir=is_dir,
                is_file=not is_dir,
                created_at=e.creationdate,
                last_modified=datetime.strptime(e.getlastmodified, "%a, %d %b %Y %H:%M:%S GMT").isoformat(),
                size=e.getcontentlength
            )
            entries.append(param)
    return entries


@contextmanager
def makedir(path):
    directory = path.absolute
    logger.info('Creating directory: %s', path.absolute)
    try:
        yield path.absolute
    except HTTPError as error:
        try:
            accept_error(error.error.response.error.msg, path=directory)
        except ResourceExistsError:
            logger.info('Directory already exists: %s', directory)
    logger.info('Directory created: %s', directory)


@contextmanager
def copy(path, destination):
    destination = destination.join(path.name).absolute
    logger.info('Copying: %s to: %s', path.absolute, destination)
    yield path.absolute, destination


@contextmanager
def move(path, destination):
    destination = destination.join(path.name).absolute
    logger.info('Moving: %s to: %s', path.absolute, destination)
    yield path.absolute, destination


def delete_generator(*paths):
    for path in paths:
        logger.info('Deleting item: %s', path.absolute)
        yield path


@contextmanager
def handle(path):
    logger.info('Getting file handle: %s', path.reference)
    yield path.absolute


@contextmanager
def handle_many(directory, objects):
    param = Object()
    param.paths = ['/'.join([directory.absolute, item]) for item in objects]
    param.snapshot = Object()
    param._classname = 'BackupRepository'  # pylint: disable=protected-access
    param.snapshot.location = 1
    param.snapshot.timestamp = None
    param.snapshot.path = None
    logger.info('Getting directory handle: %s', directory.reference)
    yield param


@contextmanager
def upload(name, destination, fd):
    fd, *_ = common.encode_stream(fd, 0)
    param = dict(
        name=name,
        fullpath=f'{destination.absolute}/{name}',
        filedata=fd
    )
    logger.info('Uploading: %s to: %s', name, destination.reference)
    yield param


file_access_errors = {
    "File exists": ResourceExistsError,
    "Creating a folder in this location is forbidden": RestrictedPathError
}


def accept_error(error_message, **kwargs):
    """
    Check if response contains an error.
    """
    if error_message not in ['OK']:
        exception_classname = file_access_errors.get(error_message, RemoteStorageError)
        try:
            if exception_classname:
                raise exception_classname(**kwargs)
        except ResourceExistsError as error:
            logger.warning('Resource already exists: a file or folder with this name already exists.')
            raise error
        except RestrictedPathError as error:
            logger.error('Creating a folder in the specified location is forbidden.')
            raise error
        except RemoteStorageError as error:
            logger.error('An error occurred while performing operation.')
            raise error



from .base_command import BaseCommand
from ...core import decorator


class Portals(BaseCommand):

    @decorator.update_current_tenant
    async def browse(self, tenant):
        """
        Browse a tenant

        :param str tenant: Name of the tenant to browse
        """
        await self._core.v1.api.put('/currentPortal', tenant)

    async def browse_global_admin(self):
        """
        Browse the Global Admin
        """
        await self.browse('')



import functools
import asyncio
import logging
import aiohttp

from yarl import URL
from ..exceptions.transport import TLSError


logger = logging.getLogger('cterasdk.http')


class Session:
    """Asynchronous HTTP Session"""

    def __init__(self, settings, trace):
        self._settings = settings
        self._trace = trace
        self._session = None

    @property
    def session(self):
        if self.closed:
            self._session = aiohttp.ClientSession(**self._settings, **self._trace)
        return self._session

    @property
    def cookies(self):
        return CookieJar(self.session.cookie_jar)

    async def request(self, r, *, await_promise=False, on_response=None):
        if await_promise:
            return await self.await_promise(r, on_response=on_response)
        return await self.promise(r, on_response=on_response)

    async def promise(self, r, *, on_response=None):
        return await self._request(r, on_response=on_response)

    async def await_promise(self, r, *, on_response=None):
        promise = await self._request(r, on_response=on_response)
        return await promise

    async def _request(self, r, *, on_response=None):
        try:
            response = await self.session.request(r.method, r.url, **r.kwargs)
            return asyncio.create_task(on_response(response))
        except aiohttp.ClientSSLError as error:
            logger.warning(error)
            raise TLSError(error.host, error.port) from error
        except aiohttp.ClientProxyConnectionError as error:
            logger.warning(error)
            raise
        except aiohttp.ClientConnectorError as error:
            logger.warning(error)
            raise ConnectionError(error)
        except aiohttp.ServerDisconnectedError as error:
            logger.warning(error)
            raise ConnectionError(error)
        except asyncio.TimeoutError as error:
            logger.debug('Request timed out while making an HTTP request.')
            raise TimeoutError(error)

    @property
    def closed(self):
        return self._session.closed if self._session is not None else True

    async def close(self):
        if self._session is not None:
            await self._session.close()


def decorate_stream_error(stream_reader):
    @functools.wraps(stream_reader)
    async def wrapper(self, n=-1):
        try:
            return await stream_reader(self, n)
        except aiohttp.ClientPayloadError as error:
            logger.warning(error)
            raise IOError(error)
    return wrapper


class CookieJar:

    def __init__(self, cookie_jar):
        self._cookie_jar = cookie_jar

    def update_cookies(self, cookies, response_url=None):
        self._cookie_jar.update_cookies(cookies, URL(response_url))

    def get(self, key):
        for cookie in self._cookie_jar:
            if cookie.key == key:
                return cookie.value
        return None

    def clear(self):
        self._cookie_jar.clear()


class BaseRequest:
    """HTTP Request"""

    def __init__(self, method, url, **kwargs):
        self.method = method
        self.url = url
        self.kwargs = BaseRequest.accept(**kwargs)

    @staticmethod
    def accept(**kwargs):
        timeout = kwargs.get('timeout', None)
        if timeout:
            logger.debug('Setting request timeout. %s', timeout)
            kwargs['timeout'] = aiohttp.ClientTimeout(**timeout)
        return kwargs


class GetRequest(BaseRequest):
    """POST"""

    def __init__(self, url, **kwargs):
        super().__init__('GET', url, **kwargs)


class PutRequest(BaseRequest):
    """PUT"""

    def __init__(self, url, **kwargs):
        super().__init__('PUT', url, **kwargs)


class PostRequest(BaseRequest):
    """POST"""

    def __init__(self, url, **kwargs):
        super().__init__('POST', url, **kwargs)


class DeleteRequest(BaseRequest):
    """DELETE"""

    def __init__(self, url, **kwargs):
        super().__init__('DELETE', url, **kwargs)


class PropfindRequest(BaseRequest):
    """PROPFIND"""

    def __init__(self, url, **kwargs):
        super().__init__('PROPFIND', url, **kwargs)


class MkcolRequest(BaseRequest):
    """MKCOL"""

    def __init__(self, url, **kwargs):
        super().__init__('MKCOL', url, **kwargs)


class CopyRequest(BaseRequest):
    """COPY"""

    def __init__(self, url, **kwargs):
        super().__init__('COPY', url, **kwargs)


class MoveRequest(BaseRequest):
    """MOVE"""

    def __init__(self, url, **kwargs):
        super().__init__('MOVE', url, **kwargs)


class FormData(aiohttp.FormData):
    """Class representing URL-encoded form data"""



import logging

from ..lib import track, ErrorStatus
from .enum import Mode, SyncStatus, Acl
from .base_command import BaseCommand
from ..common import Object, ThrottlingRule, FilterBackupSet, FileFilterBuilder


logger = logging.getLogger('cterasdk.edge')


class Sync(BaseCommand):
    """
    Edge Filer Cloud Sync APIs

    :ivar cterasdk.edge.sync.CloudSyncBandwidthThrottling throttling: Object holding the Edge Filer's bandwidth throttling APIs
    """

    def __init__(self, portal):
        super().__init__(portal)
        self.throttling = CloudSyncBandwidthThrottling(self._edge)

    def get_status(self):
        """ Retrieve the Cloud Sync status """
        return self._edge.api.get('/proc/cloudsync/serviceStatus')

    def is_disabled(self):
        """ Check if Cloud Sync is disabled """
        return self._edge.api.get('/config/cloudsync/mode') == Mode.Disabled

    def is_enabled(self):
        """ Check if Cloud Sync is enabled """
        return self._edge.api.get('/config/cloudsync/mode') == Mode.Enabled

    def suspend(self, wait=True):
        """
        Suspend Cloud Sync

        :param bool wait: Wait for synchronization to stop
        """
        logger.info("Suspending cloud sync.")
        self._edge.api.put('/config/cloudsync/mode', Mode.Disabled)
        if wait:
            self._track_status(
                [
                    SyncStatus.Off
                ],
                [
                    SyncStatus.Synced,
                    SyncStatus.Syncing,
                    SyncStatus.Scanning,
                    SyncStatus.ConnectingFolders,
                    SyncStatus.InitializingConnection
                ],
                [
                    # No Transient Status
                ],
                [
                    SyncStatus.DisconnectedPortal,
                    SyncStatus.Unlicensed,
                    SyncStatus.ServiceUnavailable,
                    SyncStatus.VolumeUnavailable,
                    SyncStatus.ShouldSupportWinNtAcl,
                    SyncStatus.InternalError,
                    SyncStatus.ClocksOutOfSync,
                    SyncStatus.NoFolder
                ]
            )
        logger.info("Cloud sync suspended.")

    def unsuspend(self):
        """ Unsuspend Cloud Sync """
        logger.info("Unsuspending cloud sync.")
        self._edge.api.put('/config/cloudsync/mode', Mode.Enabled)
        try:
            self._track_status(
                [
                    SyncStatus.Synced,
                    SyncStatus.Syncing,
                    SyncStatus.Scanning,
                    SyncStatus.UpgradingDataBase
                ],
                [
                    SyncStatus.ConnectingFolders,
                    SyncStatus.InitializingConnection
                ],
                [
                    SyncStatus.DisconnectedPortal,
                    SyncStatus.Unlicensed,
                    SyncStatus.Off
                ],
                [
                    SyncStatus.ServiceUnavailable,
                    SyncStatus.VolumeUnavailable,
                    SyncStatus.ShouldSupportWinNtAcl,
                    SyncStatus.InternalError,
                    SyncStatus.ClocksOutOfSync,
                    SyncStatus.NoFolder
                ]
            )
            logger.info('Unsuspended cloud sync.')
        except ErrorStatus as error:
            if error.status == SyncStatus.ShouldSupportWinNtAcl:
                logger.warning('Windows ACL enabled folder cannot be synchronized to this device.')
                self._edge.api.put('/config/fileservices/share/cloud/access', Acl.WindowsNT)
                logger.info('Updated network share access. %s', {'share': 'cloud', 'access': Acl.WindowsNT})
            else:
                logger.error("An error occurred while unsuspendeding sync. %s", {'status': error.status})

    def _track_status(self, success, progress, transient, failure):
        track(self._edge, '/proc/cloudsync/serviceStatus/id', success, progress, transient, failure)

    def refresh(self):
        """ Refresh Cloud Folders """
        logger.info("Refreshing cloud folders.")
        self._edge.api.execute("/config/cloudsync/cloudExtender", "refreshPaths", None)
        logger.info("Completed refreshing cloud folders.")

    def exclude_files(self, extensions=None, filenames=None, paths=None, custom_exclusion_rules=None):
        """
        Exclude files from Cloud Sync. This method will override any existing file exclusion rules
        Use :func:`cterasdk.common.types.FileFilterBuilder` to build custom file exclusion rules`

        :param list[str] extensions: List of file extensions
        :param list[str] filenames: List of file names
        :param list[str] paths: List of file paths
        :param list[cterasdk.common.types.FilterBackupSet] rules: Set of custom exclusion rules
        """
        rules = []
        if extensions:
            param = FilterBackupSet('List of file extensions to exclude from sync',
                                    filter_rules=[FileFilterBuilder.extensions().include(extensions).build()])
            rules.append(param)
        if filenames:
            param = FilterBackupSet('List of file names to exclude from sync',
                                    filter_rules=[FileFilterBuilder.names().include(filenames).build()])
            rules.append(param)
        if paths:
            param = FilterBackupSet('List of file paths to exclude from sync',
                                    filter_rules=[FileFilterBuilder.paths().include(filenames).build()])
            rules.append(param)
        if custom_exclusion_rules:
            rules.extend(custom_exclusion_rules)

        if rules:
            logger.info('Setting sync exclusion rules')
            self._edge.api.put('/config/cloudsync/excludeFiles', rules)
            logger.info('Sync exclusion rules set')

    def remove_file_exclusion_rules(self):
        """
        Remove previously configured sync exclusion rules
        """
        logger.info('Removing sync exclusion rules')
        self._edge.api.put('/config/cloudsync/excludeFiles', None)
        logger.info('Sync exclusion rules removed')

    def get_linux_avoid_using_fanotify(self):
        logger.info('Getting LinuxAvoidUsingFAnotify')
        return self._edge.api.get('/config/cloudsync/LinuxAvoidUsingFAnotify')

    def set_linux_avoid_using_fanotify(self, avoid):
        logger.info('Setting LinuxAvoidUsingFAnotify to %s', avoid)
        self._edge.api.put('/config/cloudsync/LinuxAvoidUsingFAnotify', avoid)

    def evict(self, path, wait=False):
        """
        Evict a directory from the Edge Filer

        :param str path: Directory path
        :param bool wait: Wait for eviction task to complete, defaults to ``False``
        :returns: Task status object, or an awaitable task object
        :rtype: cterasdk.common.object.Object or :class:`cterasdk.lib.tasks.AwaitableEdgeTask`
        """
        param = Object()
        param._classname = 'evictFolderParam'  # pylint: disable=protected-access
        param.path = path
        ref = self._edge.api.execute('/config/cloudsync', 'evictFolder', param)
        if wait:
            return self._edge.tasks.wait(ref)
        return self._edge.tasks.awaitable_task(ref)


class CloudSyncBandwidthThrottling(BaseCommand):
    """ Edge Filer Cloud Sync Bandwidth Throttling APIs """

    def get_policy(self):
        """
        Get the bandwidth throttling policy

        :returns: a list of bandwidth throttling rules
        :rtype: list[cterasdk.common.types.ThrottlingRule]
        """
        rules = self._edge.api.get('/config/cloudsync/syncThrottlingTopic/multiThrottling') or []
        return [ThrottlingRule.from_server_object(rule) for rule in rules]

    def set_policy(self, rules):
        """
        Set the bandwidth throttling policy

        :param list[cterasdk.common.types.ThrottlingRule] rules: List of bandwidth throttling rules
        """
        return self._edge.api.put('/config/cloudsync/syncThrottlingTopic/multiThrottling', [rule.to_server_object() for rule in rules])
